/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include "mrp-intl.h"
#include "mrp-task.h"
#include "mrp-resource.h"
#include "mrp-marshal.h"
#include "mrp-assignment.h"

struct _MrpAssignmentPriv {
	MrpTask     *task;
	MrpResource *resource;

	gint         units;
};

/* Properties */
enum {
        PROP_0,
        PROP_TASK,
        PROP_RESOURCE,
        PROP_UNITS
};


static void mpa_class_init        (MrpAssignmentClass *klass);
static void mpa_init              (MrpAssignment      *assignment);
static void mpa_finalize          (GObject            *object);
static void mpa_set_property      (GObject            *object,
				   guint               prop_id,
				   const GValue       *value,
				   GParamSpec         *pspec);
static void mpa_get_property      (GObject            *object,
				   guint               prop_id,
				   GValue             *value,
				   GParamSpec         *pspec);
/*static void mpa_object_removed_cb (MrpObject          *object,
  MrpAssignment      *assignment);*/


static MrpObjectClass *parent_class;

GType
mrp_assignment_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (MrpAssignmentClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) mpa_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (MrpAssignment),
			0,              /* n_preallocs */
			(GInstanceInitFunc) mpa_init,
		};

		type = g_type_register_static (MRP_TYPE_OBJECT, 
					       "MrpAssignment", 
					       &info, 0);
	}

	return type;
}

static void
mpa_class_init (MrpAssignmentClass *klass)
{
        GObjectClass   *object_class     = G_OBJECT_CLASS (klass);
        
        parent_class = MRP_OBJECT_CLASS (g_type_class_peek_parent (klass));
        
        object_class->finalize     = mpa_finalize;
        object_class->set_property = mpa_set_property;
        object_class->get_property = mpa_get_property;

	/* Properties */
        g_object_class_install_property (object_class,
                                         PROP_TASK,
                                         g_param_spec_object ("task",
							      _("Task with assignment"),
							      NULL,
							      MRP_TYPE_TASK,
							      G_PARAM_READWRITE));
        g_object_class_install_property (object_class,
                                         PROP_RESOURCE,
                                         g_param_spec_object ("resource",
							      _("Allocated resource"),
							      NULL,
							      MRP_TYPE_RESOURCE,
							      G_PARAM_READWRITE));
        g_object_class_install_property (object_class,
                                         PROP_UNITS,
                                         g_param_spec_int ("units",
							   _("Allocated units"),
							   NULL,
							   -1, 
							   G_MAXINT,
							   0,
							   G_PARAM_READWRITE));
}

static void
mpa_init (MrpAssignment *assignment)
{
        MrpAssignmentPriv *priv;
        
        priv = g_new0 (MrpAssignmentPriv, 1);
        
        assignment->priv = priv;
}

static void
mpa_finalize (GObject *object)
{
        MrpAssignment     *assignment = MRP_ASSIGNMENT (object);
        MrpAssignmentPriv *priv;
        
        priv = assignment->priv;

	if (priv->task) {
		g_object_unref (priv->task);
		priv->task = NULL;
	}

	if (priv->resource) {
		g_object_unref (priv->resource);
		priv->resource = NULL;
	}

        if (G_OBJECT_CLASS (parent_class)->finalize) {
                (* G_OBJECT_CLASS (parent_class)->finalize) (object);
        }
}

static void
mpa_set_property (GObject        *object,
                  guint           prop_id,
                  const GValue   *value,
                  GParamSpec     *pspec)
{
	MrpAssignment     *assignment;
	MrpAssignmentPriv *priv;
	
	g_return_if_fail (MRP_IS_ASSIGNMENT (object));
	
	assignment = MRP_ASSIGNMENT (object);
	priv       = assignment->priv;
	
	switch (prop_id) {
	case PROP_TASK:
		if (priv->task) {
			g_object_unref (priv->task);
			priv->task = NULL;
		}

		priv->task = MRP_TASK (
			g_object_ref (g_value_get_object (value)));
		break;
	case PROP_RESOURCE:
		if (priv->resource) {
			g_object_unref (priv->resource);
		}

		priv->resource = MRP_RESOURCE (
			g_object_ref (g_value_get_object (value)));
		break;
	case PROP_UNITS:
		priv->units = g_value_get_int (value);
		break;
	default:
		break;
	}
}

static void
mpa_get_property (GObject      *object,
                  guint         prop_id,
                  GValue       *value,
                  GParamSpec   *pspec)
{
	MrpAssignment     *assignment;
	MrpAssignmentPriv *priv;
	
	g_return_if_fail (MRP_IS_ASSIGNMENT (object));
	
	assignment = MRP_ASSIGNMENT (object);
	priv       = assignment->priv;
	
	switch (prop_id) {
	case PROP_TASK:
		g_value_set_object (value, priv->task);
		break;
	case PROP_RESOURCE:
		g_value_set_object (value, priv->resource);
		break;
	case PROP_UNITS:
		g_value_set_int (value, priv->units);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

#if 0
static void
mpa_object_removed_cb (MrpObject *object, MrpAssignment *assignment)
{
	g_return_if_fail (MRP_IS_OBJECT (object));
	g_return_if_fail (MRP_IS_ASSIGNMENT (assignment));

	mrp_object_removed (MRP_OBJECT (assignment));
}
#endif

MrpAssignment *
mrp_assignment_new (void)
{
        MrpAssignment *assignment;
        
        assignment = g_object_new (MRP_TYPE_ASSIGNMENT, NULL);
        
        return assignment;
}

MrpTimeInterval *
mrp_assignment_get_interval (MrpAssignment *assignment)
{
	g_return_val_if_fail (MRP_IS_ASSIGNMENT (assignment), NULL);
	
	return mrp_task_get_interval (assignment->priv->task);
}

MrpTask *
mrp_assignment_get_task (MrpAssignment *assignment)
{
	g_return_val_if_fail (MRP_IS_ASSIGNMENT (assignment), NULL);
	
	return assignment->priv->task;
}

MrpResource *
mrp_assignment_get_resource (MrpAssignment *assignment)
{
	g_return_val_if_fail (MRP_IS_ASSIGNMENT (assignment), NULL);
	
	return assignment->priv->resource;
}

gint
mrp_assignment_get_units (MrpAssignment *assignment)
{
	g_return_val_if_fail (MRP_IS_ASSIGNMENT (assignment), -1);
	
	return assignment->priv->units;
}
