/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include "mrp-intl.h"
#include "mrp-private.h"
#include "mrp-calendar.h"

/* Properties */
enum {
	PROP_0,
	PROP_NAME
};

struct _MrpCalendarPriv {
	gchar          *name;

	MrpCalendar    *base_calendar;

	/* This can override the default calendar */
	MrpCalendarDay *default_days[7];

	/* This can override single days and is hashed on the date */
	GHashTable     *days;

	/* Temporary!! */
	MrpCalendarDay *working_day;
	MrpCalendarDay *nonworking_day;
};

struct _MrpCalendarDay {
	gchar *name;
	
	GList *intervals;

	guint  ref_count;
};

static void           calendar_class_init         (MrpCalendarClass *class);
static void           calendar_init               (MrpCalendar      *module);
static void           calendar_finalize           (GObject          *object);
static void           calendar_set_property       (GObject          *object,
						   guint             prop_id,
						   const GValue     *value,
						   GParamSpec       *pspec);
static void           calendar_get_property       (GObject          *object,
						   guint             prop_id,
						   GValue           *value,
						   GParamSpec       *pspec);

MrpCalendarDay *      calendar_get_day            (MrpCalendar      *calendar,
						   mrptime           time);
MrpCalendarDay *      calendar_get_default_day    (MrpCalendar      *calendar,
						   mrptime           time);
void                  calendar_day_free           (MrpCalendarDay   *day);

/* Temporary!! */
void                  calendar_setup_hardcoded    (MrpCalendar      *calendar);

	
static MrpObjectClass *parent_class;


GType
mrp_calendar_get_type (void)
{
	static GType object_type = 0;

	if (!object_type) {
		static const GTypeInfo object_info = {
			sizeof (MrpCalendarClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) calendar_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (MrpCalendar),
			0,              /* n_preallocs */
			(GInstanceInitFunc) calendar_init,
		};

		object_type = g_type_register_static (MRP_TYPE_OBJECT, 
						      "MrpCalendar", 
						      &object_info, 0);
	}

	return object_type;
}

static void
calendar_class_init (MrpCalendarClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = MRP_OBJECT_CLASS (g_type_class_peek_parent (klass));

	object_class->finalize = calendar_finalize;
	object_class->set_property = calendar_set_property;
	object_class->get_property = calendar_get_property;
	
	g_object_class_install_property (object_class,
					 PROP_NAME,
					 g_param_spec_string ("name",
							      _("Calendar Name"),
							      NULL,
							      "empty",
							      G_PARAM_READWRITE));
}

static void
calendar_init (MrpCalendar *calendar)
{
	MrpCalendarPriv *priv;
	
	priv = g_new0 (MrpCalendarPriv, 1);
	
	priv->name          = NULL;
	priv->base_calendar = NULL;
	priv->days          = g_hash_table_new (NULL, NULL);
	
	calendar->priv = priv;

	calendar_setup_hardcoded (calendar);
}

static void
calendar_finalize (GObject *object)
{
	MrpCalendar     *calendar;
	MrpCalendarPriv *priv;

	calendar = MRP_CALENDAR (object);
	priv     = calendar->priv;

	g_free (calendar->priv);
	calendar->priv = NULL;

	if (G_OBJECT_CLASS (parent_class)->finalize) {
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
	}
}

static void
calendar_set_property (GObject         *object,
		       guint            prop_id,
		       const GValue    *value,
		       GParamSpec      *pspec)
{
	MrpCalendar *calendar;

	calendar = MRP_CALENDAR (object);
	
	switch (prop_id) {
	case PROP_NAME:
		calendar->priv->name = g_strdup (g_value_get_string (value));
		break;

	default:
		break;
	}
}

static void
calendar_get_property (GObject    *object,
		       guint       prop_id,
		       GValue     *value,
		       GParamSpec *pspec)
{
	MrpCalendar *calendar;

	calendar = MRP_CALENDAR (object);

	switch (prop_id) {
	case PROP_NAME:
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

MrpCalendarDay * 
calendar_get_day (MrpCalendar *calendar, mrptime time)
{
        MrpCalendarPriv *priv;
        MrpCalendarDay  *ret_val;
	gconstpointer    key;

        g_return_val_if_fail (MRP_IS_CALENDAR (calendar), NULL);

        priv = calendar->priv;

	key = GINT_TO_POINTER ((int) mrp_time_align_day (time));
	
        ret_val = g_hash_table_lookup (priv->days, key);

        if (!ret_val && priv->base_calendar) {
                return calendar_get_day (calendar, time);
        }

        return ret_val;
}

MrpCalendarDay * 
calendar_get_default_day (MrpCalendar *calendar, mrptime time)
{
        MrpCalendarPriv *priv;
        MrpCalendarDay  *ret_val;
	gint             day;
	
        g_return_val_if_fail (MRP_IS_CALENDAR (calendar), NULL);

        priv = calendar->priv;

	day = mrp_time_day_of_week (time);

	ret_val = priv->default_days[day];
	
	if (!ret_val && priv->base_calendar) {
		return calendar_get_default_day (priv->base_calendar, time);
	}

	return ret_val;
}

void
calendar_day_free (MrpCalendarDay *day)
{
	g_return_if_fail (day != NULL);

	g_free (day->name);

	if (day->intervals) {
		g_list_foreach (day->intervals, 
				(GFunc) mrp_interval_unref,
				NULL);
		g_list_free (day->intervals);
	}
}

void
calendar_setup_hardcoded (MrpCalendar *calendar)
{
	MrpCalendarPriv *priv;
 	MrpInterval     *interval;
	gint             i;

	g_return_if_fail (MRP_IS_CALENDAR (calendar));
	
	priv = calendar->priv;

	priv->working_day = mrp_calendar_day_new ();
	priv->nonworking_day = mrp_calendar_day_new ();
	
	interval = mrp_interval_new (0, 24*60*60 - 0, 
				     MRP_INTERVAL_TYPE_NONWORK);

	priv->nonworking_day->intervals = g_list_prepend (NULL, interval);
	
	interval = mrp_interval_new (0, 8*60*60 - 0,
				     MRP_INTERVAL_TYPE_NONWORK);
	
	priv->working_day->intervals = 
		g_list_append (priv->working_day->intervals, interval);
	
	interval = mrp_interval_new (8*60*60, 12*60*60 - 0,
				     MRP_INTERVAL_TYPE_WORK);

	priv->working_day->intervals = 
		g_list_append (priv->working_day->intervals, interval);

	interval = mrp_interval_new (12*60*60, 13*60*60 - 0,
				     MRP_INTERVAL_TYPE_NONWORK);
	
	priv->working_day->intervals = 
		g_list_append (priv->working_day->intervals, interval);

	interval = mrp_interval_new (13*60*60, 17*60*60 - 0,
				     MRP_INTERVAL_TYPE_WORK);

	priv->working_day->intervals = 
		g_list_append (priv->working_day->intervals, interval);

	interval = mrp_interval_new (17*60*60, 24*60*60 - 0,
				     MRP_INTERVAL_TYPE_NONWORK);

	priv->working_day->intervals = 
		g_list_append (priv->working_day->intervals, interval);

	for (i = 1; i < 6; ++i) {
		mrp_calendar_set_default_day (calendar, priv->working_day, i);
	}
	
	mrp_calendar_set_default_day (calendar, priv->nonworking_day, 0);
	mrp_calendar_set_default_day (calendar, priv->nonworking_day, 6);
}

MrpCalendar *
mrp_calendar_new (void)
{
	MrpCalendar *calendar;

	calendar = g_object_new (MRP_TYPE_CALENDAR, NULL);

	return calendar;
}

MrpCalendar *
mrp_calendar_copy (MrpCalendar *calendar)
{
	MrpCalendar *ret_val;
	
	ret_val = mrp_calendar_new ();

	return ret_val;
}

MrpCalendar *
mrp_calendar_derive (MrpCalendar *parent)
{
	MrpCalendar *ret_val;
	
	ret_val = mrp_calendar_new ();
	ret_val->priv->base_calendar = parent;
	
	return ret_val;
}

/* Manipulating */
void      
mrp_calendar_set_day (MrpCalendar    *calendar,
		      MrpCalendarDay *day,
		      mrptime         time)
{
	MrpCalendarPriv *priv;
	MrpCalendarDay  *old_day;
	gpointer         key;
	
	g_return_if_fail (MRP_IS_CALENDAR (calendar));
	g_return_if_fail (day != NULL);
	
	priv = calendar->priv;
	
	key = GINT_TO_POINTER ((int) mrp_time_align_day (time));

	old_day = g_hash_table_lookup (priv->days, key);
	
	if (old_day) {
		g_hash_table_steal (priv->days, key);
		mrp_calendar_day_unref (old_day);
	}

	g_hash_table_insert (priv->days, key, mrp_calendar_day_ref (day));
}

void
mrp_calendar_set_default_day (MrpCalendar    *calendar,
			      MrpCalendarDay *day,
			      gint            day_of_week)
{
	MrpCalendarPriv *priv;
	MrpCalendarDay  *old_day;
	
	g_return_if_fail (MRP_IS_CALENDAR (calendar));
	g_return_if_fail (day != NULL);

	priv = calendar->priv;

	old_day = priv->default_days[day_of_week];

	if (old_day) {
		mrp_calendar_day_unref (old_day);
	}

	priv->default_days[day_of_week] = mrp_calendar_day_ref (day);
}

/* Querying */
MrpCalendarDay * 
mrp_calendar_get_day (MrpCalendar *calendar, mrptime time)
{
        MrpCalendarPriv *priv;
        MrpCalendarDay  *ret_val;
	gconstpointer    key;

        g_return_val_if_fail (MRP_IS_CALENDAR (calendar), NULL);

        priv = calendar->priv;

	key = GINT_TO_POINTER ((int) mrp_time_align_day (time));
	
        ret_val = g_hash_table_lookup (priv->days, key);

        if (!ret_val && priv->base_calendar) {
                ret_val = calendar_get_day (priv->base_calendar, time);
        }

        if (!ret_val) {
                ret_val = mrp_calendar_get_default_day (calendar, time);
        }

        return ret_val;
}

MrpCalendarDay *
mrp_calendar_get_default_day (MrpCalendar *calendar, mrptime time)
{
        MrpCalendarPriv *priv;
        gint             day;
        MrpCalendarDay  *ret_val;

        /* Look up time in priv->default_days, if it's there, return it */
        /* Otherwise, return mrp_calendar_get_default_day (base_cal, time); */

        g_return_val_if_fail (MRP_IS_CALENDAR (calendar), NULL);

        priv = calendar->priv;

        day = mrp_time_day_of_week (time);

        ret_val = priv->default_days[day];

        if (!ret_val && priv->base_calendar) {
                return mrp_calendar_get_default_day (calendar, time);
        }

        return ret_val;
}

MrpCalendarDay *
mrp_calendar_day_new (void)
{
	MrpCalendarDay *ret_val;
	
	ret_val = g_new0 (MrpCalendarDay, 1);
	
	ret_val->name      = g_strdup ("");
	ret_val->intervals = NULL;
	ret_val->ref_count = 1;

	return ret_val;
}

MrpCalendarDay *
mrp_calendar_day_ref (MrpCalendarDay *day)
{
	g_return_val_if_fail (day != NULL, NULL);
	
	day->ref_count++;

	return day;
}

void
mrp_calendar_day_unref (MrpCalendarDay *day)
{
	g_return_if_fail (day != NULL);
	
	day->ref_count--;
	
	if (day->ref_count <= 0) {
		calendar_day_free (day);
	}
}

GList *
mrp_calendar_get_intervals (MrpCalendarDay *day)
{
	g_return_val_if_fail (day != NULL, NULL);
	
	return day->intervals;
}

