/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002 CodeFactory AB
 * Copyright (C) 2001-2002 Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002 Mikael Hallendal <micke@codefactory.se>
 * Copyright (C) 2001-2002 Alvaro del Castillo <acs@barrapunto.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <string.h>
#include "mrp-intl.h"
#include "mrp-marshal.h"
#include "mrp-task-manager.h"
#include "mrp-storage-module-factory.h"
#include "mrp-storage-module.h"
#include "mrp-private.h"
#include "mrp-time.h"
#include "mrp-property.h"
#include "mrp-project.h"

struct _MrpProjectPriv {
	gchar            *uri;
	gchar            *storage_method;

	MrpTaskManager   *task_manager;

	GList            *resources;
	GList            *groups;
	
	MrpStorageModule *primary_storage;

	mrptime           project_start;

	gchar            *organization;
	gchar            *manager;
	gchar            *name;
	
	/*GList          *storage_modules;*/        /* <MrpStorageModule> */
	gboolean          needs_saving;
	gboolean          untouched;
	
	MrpGroup         *default_group;
	
	/* Property stuff */
	GParamSpecPool   *property_pool;

	MrpCalendar      *calendar;
};

/* Properties */
enum {
	PROP_0,
	PROP_PRIMARY_STORAGE,
	PROP_PROJECT_START,
	PROP_NAME,
	PROP_ORGANIZATION,
	PROP_MANAGER,
	PROP_DEFAULT_GROUP
};

/* Signals */
enum {
	LOADED,
	RESOURCE_ADDED,
	RESOURCE_REMOVED,
	GROUP_ADDED,
	GROUP_REMOVED,
	DEFAULT_GROUP_CHANGED,
	TASK_INSERTED,
	TASK_REMOVED,
	TASK_MOVED,
	NEEDS_SAVING_CHANGED,
	PROPERTY_ADDED,
	PROPERTY_CHANGED,
	PROPERTY_REMOVED,
	LAST_SIGNAL
};


static void   mpp_class_init                (MrpProjectClass  *klass);
static void   mpp_init                      (MrpProject       *project);
static void   mpp_finalize                  (GObject          *object);
static void   mpp_set_property              (GObject          *object,
					     guint             prop_id,
					     const GValue     *value,
					     GParamSpec       *pspec);
static void   mpp_get_property              (GObject          *object,
					     guint             prop_id,
					     GValue           *value,
					     GParamSpec       *pspec);
static void   mpp_connect_object            (MrpObject        *object,
					     MrpProject       *project);
static gchar *mpp_uri_get_storage_string    (const gchar      *uri);


static GObjectClass *parent_class;
static guint signals[LAST_SIGNAL];

GType
mrp_project_get_type (void)
{
	static GType object_type = 0;

	if (!object_type) {
		static const GTypeInfo object_info = {
			sizeof (MrpProjectClass),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) mpp_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof (MrpProject),
			0,              /* n_preallocs */
			(GInstanceInitFunc) mpp_init,
		};

		object_type = g_type_register_static (G_TYPE_OBJECT, 
						      "MrpProject", 
						      &object_info, 0);
	}

	return object_type;
}

static void
mpp_class_init (MrpProjectClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize     = mpp_finalize;
	object_class->set_property = mpp_set_property;
	object_class->get_property = mpp_get_property;

	signals[LOADED] = g_signal_new
		("loaded",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */ 
		 NULL, NULL,
		 mrp_marshal_VOID__VOID,
		 G_TYPE_NONE, 0);

	signals[RESOURCE_ADDED] = g_signal_new
		("resource_added",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_RESOURCE);

	signals[RESOURCE_REMOVED] = g_signal_new
		("resource_removed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_RESOURCE);
	
	signals[GROUP_ADDED] = g_signal_new
		("group_added",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_GROUP);

	signals[GROUP_REMOVED] = g_signal_new
		("group_removed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_GROUP);

	signals[DEFAULT_GROUP_CHANGED] = g_signal_new
		("default_group_changed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0, /*G_STRUCT_OFFSET (MrpProjectClass, method), */
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_GROUP);
	
	signals[TASK_INSERTED] = g_signal_new
		("task_inserted",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_TASK);

	signals[TASK_REMOVED] = g_signal_new
		("task_removed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_TASK);

	signals[TASK_MOVED] = g_signal_new
		("task_moved",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__OBJECT,
		 G_TYPE_NONE,
		 1, MRP_TYPE_TASK);

	signals[NEEDS_SAVING_CHANGED] = g_signal_new
		("needs_saving_changed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__BOOLEAN,
		 G_TYPE_NONE,
		 1, G_TYPE_BOOLEAN);

	signals[PROPERTY_ADDED] = g_signal_new
		("property_added",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__POINTER,
		 G_TYPE_NONE,
		 1, G_TYPE_POINTER);
		
	signals[PROPERTY_CHANGED] = g_signal_new
		("property_changed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__POINTER,
		 G_TYPE_NONE,
		 1, G_TYPE_POINTER);
		
	signals[PROPERTY_REMOVED] = g_signal_new
		("property_removed",
		 G_TYPE_FROM_CLASS (klass),
		 G_SIGNAL_RUN_LAST,
		 0,
		 NULL, NULL,
		 mrp_marshal_VOID__POINTER,
		 G_TYPE_NONE,
		 1, G_TYPE_POINTER);

	/* Properties. */
	g_object_class_install_property (object_class,
					 PROP_PRIMARY_STORAGE,
					 g_param_spec_object ("primary-storage",
							      _("Primary Storage Module"),
							      _("The primary storage module to use for the project"),
							      G_TYPE_OBJECT,
							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	g_object_class_install_property (object_class,
					 PROP_PROJECT_START,
					 mrp_param_spec_time ("project-start",
							      _("Project start"),
							      _("The start date of the project"),
							      G_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_NAME,
					 g_param_spec_string ("name",
							   _("Project Name"),
							   NULL,
							   "",
							   G_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_ORGANIZATION,
					 g_param_spec_string ("organization",
							   _("Organization"),
							   NULL,
							   "",
							   G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_MANAGER,
					 g_param_spec_string ("manager",
							   _("Project Manager"),
							   NULL,
							   "",
							   G_PARAM_READWRITE));

	g_object_class_install_property (object_class,
					 PROP_DEFAULT_GROUP,
					 g_param_spec_object ("default-group",
							      _("Default Group"),
							      _("Default group for new resources"),
							      MRP_TYPE_GROUP,
							      G_PARAM_READWRITE));
}

static void
mpp_init (MrpProject *project)
{
	MrpProjectPriv *priv;
	MrpProperty    *property;
	
	project->priv = g_new0 (MrpProjectPriv, 1);

	priv = project->priv;
	
	priv->needs_saving  = FALSE;
	priv->untouched     = TRUE;
	priv->project_start = mrp_time_align_day (mrp_time_current_time ());
	priv->resources     = NULL;
	priv->groups        = NULL;
	priv->organization  = g_strdup ("");;
	priv->manager       = g_strdup ("");;
	priv->name          = g_strdup ("");
	priv->calendar      = mrp_calendar_new ();

	priv->property_pool = g_param_spec_pool_new (TRUE);
	
        /* Move these into the cost plugin */
	property = mrp_property_new ("cost", 
				     MRP_PROPERTY_TYPE_COST,
				     "Cost",
				     "standard cost for a resource");
	mrp_project_add_property (project,
				  MRP_TYPE_RESOURCE,
				  property,
                                  FALSE);

	/* FIXME: Add back this when it's used. */
/*	property = mrp_property_new ("cost_overtime", 
				     MRP_PROPERTY_TYPE_COST,
				     "Overtime Cost",
				     "overtime cost for a resource");
	mrp_project_add_property (project,
				  MRP_TYPE_RESOURCE,
				  property,
                                  FALSE);		
*/

	priv->default_group = NULL;
}

static void
mpp_finalize (GObject *object)
{
	MrpProject *project = MRP_PROJECT (object);

	g_object_unref (project->priv->primary_storage);
	g_object_unref (project->priv->task_manager);
	
	g_free (project->priv->uri);
	g_free (project->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize) {
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
	}
}

static void
mpp_set_property (GObject      *object,
		  guint         prop_id,
		  const GValue *value,
		  GParamSpec   *pspec)
{
	MrpProject     *project;
	MrpProjectPriv *priv;
	MrpGroup       *group;
	
	project = MRP_PROJECT (object);
	priv    = project->priv;
	
	switch (prop_id) {
	case PROP_PRIMARY_STORAGE:
		priv->primary_storage = g_value_get_object (value);
		if (priv->primary_storage) {
			g_object_ref (priv->primary_storage);
		}
		
		priv->task_manager = mrp_task_manager_new (project);
		break;

	case PROP_PROJECT_START:
		priv->project_start = g_value_get_long (value);
		imrp_project_set_needs_saving (project, TRUE);
		break;

	case PROP_ORGANIZATION:
		g_free (priv->organization);
		priv->organization = g_strdup (g_value_get_string (value));
		imrp_project_set_needs_saving (project, TRUE);
		break;

	case PROP_MANAGER:
		g_free (priv->manager);
		priv->manager = g_strdup (g_value_get_string (value));
		imrp_project_set_needs_saving (project, TRUE);
		break;

	case PROP_NAME:
		g_free (priv->name);
		priv->name = g_strdup (g_value_get_string (value));
		imrp_project_set_needs_saving (project, TRUE);
		break;

	case PROP_DEFAULT_GROUP:
		group = g_value_get_object (value);
		if (priv->default_group == group) {
			break;
		}
		
		if (priv->default_group) {
			g_object_unref (priv->default_group);
		}
		
		priv->default_group = group;

		if (priv->default_group) {
			g_object_ref (priv->default_group);
		}

		g_signal_emit (project,
			       signals[DEFAULT_GROUP_CHANGED], 
			       0,
			       priv->default_group);
		imrp_project_set_needs_saving (project, TRUE);
		break;

	default:
		break;
	}
}

static void
mpp_get_property (GObject    *object,
		  guint       prop_id,
		  GValue     *value,
		  GParamSpec *pspec)
{
	MrpProject     *project;
	MrpProjectPriv *priv;
	
	project = MRP_PROJECT (object);
	priv    = project->priv;
	
	switch (prop_id) {
	case PROP_PRIMARY_STORAGE:
		g_value_set_object (value, priv->primary_storage);
		break;

	case PROP_PROJECT_START:
		g_value_set_long (value, priv->project_start);
		break;

	case PROP_ORGANIZATION:
		g_value_set_string (value, priv->organization);
		break;

	case PROP_NAME:
		g_value_set_string (value, priv->name);
		break;

	case PROP_MANAGER:
		g_value_set_string (value, priv->manager);
		break;

	case PROP_DEFAULT_GROUP:
		g_value_set_object (value, priv->default_group);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break; 
	}
}

static void
mpp_connect_object (MrpObject *object, MrpProject *project)
{
	g_object_set (G_OBJECT (object), "project", project, NULL);
}

static gchar *
mpp_uri_get_storage_string (const gchar *uri)
{
	const gchar *p;
	gchar       *method, *ret;

	if (uri == NULL) {
		return NULL;
	}
	
	for (p = uri; g_ascii_isalnum (*p) || *p == '+' || *p == '-' || *p == '.'; p++) {
		;
	}

	if (*p == ':') {
		/* Found toplevel method specification. */
		method = g_strndup (uri, p - uri);
		ret = g_ascii_strdown (method, -1);
		g_free (method);
	} else {
		ret = g_strdup ("file");
	}

	return ret;
}

MrpProject *
mrp_project_new (const gchar *primary_storage)
{
	MrpProject              *project;
	MrpStorageModuleFactory *factory;
	MrpStorageModule        *module;

	factory = mrp_storage_module_factory_get (primary_storage);
	if (factory == NULL) {
		g_error ("FIXME: Could not find storage module for method %s.",
			 primary_storage);
	}
	
	module = mrp_storage_module_factory_create_module (factory);
	if (module == NULL) {
		g_error ("FIXME: Could not create module for method %s.",
			 primary_storage);
	}

	g_type_module_unuse (G_TYPE_MODULE (factory));
	
	project = g_object_new (MRP_TYPE_PROJECT,
				"primary_storage", module,
				NULL);

	imrp_storage_module_set_project (module, project);
	
	g_object_unref (module);

	return MRP_PROJECT (project);
}

MrpProject *
mrp_project_open (const gchar *uri, GError **error)
{
	MrpProject *project;
	gchar      *storage;
	
	storage = mpp_uri_get_storage_string (uri);
	
	if (!storage) {
		g_warning ("FIXME: report error on bad URI");
		return NULL;
	}

	project = mrp_project_new (storage);

	g_free (storage);
	
	if (!mrp_project_load (project, uri, error)) {
		g_object_unref (project);
		return NULL;
	}
	
	return project;
}

/* FIXME: should we separate the uri string from the storage string?
 * E.g. storage = "mrproject-1", and uri = "http://foo.mrproject".
 * Makes sense...
 */
gboolean
mrp_project_load (MrpProject   *project,
		  const gchar  *uri,
		  GError      **error)
{
	MrpProjectPriv *priv;

	g_return_val_if_fail (MRP_IS_PROJECT (project), FALSE);
	g_return_val_if_fail (uri != NULL, FALSE);

	priv = project->priv;

	/* FIXME: this should be in a primary-storage interface on the module,
	 * since only modules that support that interface should be able to
	 * load a project directly.
	 */
	if (!mrp_storage_module_load (priv->primary_storage, uri, error)) {
		return FALSE;
	}

	if (priv->uri == NULL || strcmp (uri, priv->uri) != 0) {
		g_free (priv->uri);
		priv->uri = g_strdup (uri);
	}

	g_signal_emit (project, signals[LOADED], 0, NULL);

	imrp_project_set_needs_saving (project, FALSE);

	return TRUE;
}

gboolean
mrp_project_save (MrpProject *project, GError **error)
{
	MrpProjectPriv *priv;
	gboolean        success;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), FALSE);
	
	priv = project->priv;

	if (!priv->needs_saving) {
		return TRUE;
	}
	
	/* Sanity check the URI. */
	if (priv->uri == NULL) {
		return FALSE;
	}

	/* FIXME: this should be in a primary-storage interface on the module,
	 * since only modules that support that interface should be able to
	 * save a project directly.
	 */
	success = mrp_storage_module_save (priv->primary_storage, 
					   priv->uri, TRUE, error);

	if (!success) {
		return FALSE;
	}

	imrp_project_set_needs_saving (project, FALSE);

	return TRUE;
}

gboolean
mrp_project_save_as (MrpProject   *project,
		     const gchar  *uri,
		     gboolean      force,
		     GError      **error)
{
	MrpProjectPriv *priv;
	gboolean        success;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), FALSE);
	
	priv = project->priv;

	/* FIXME: this should be in a primary-storage interface on the module,
	 * since only modules that support that interface should be able to
	 * save a project directly.
	 */
	success = mrp_storage_module_save (priv->primary_storage, 
					   uri, force, error);
	
	if (!success) {
		/* This means that the new file exists already, or that we
		 * couldn't write to it for some reason.
		 */
		return FALSE;
	}

	g_free (priv->uri);
	priv->uri = g_strdup (uri);

	imrp_project_set_needs_saving (project, FALSE);
	
	return TRUE;
}

void
mrp_project_close (MrpProject *project)
{
	/* FIXME: impl close. */
}

MrpTaskManager *
mrp_project_get_task_manager (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return project->priv->task_manager;
}

/*MrpCalendarManager *
mrp_project_get_calendar_manager (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return project->priv->calendar_manager;
}*/

GList *
mrp_project_get_resources (MrpProject *project)
{
	GList *resources;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	resources = g_list_copy (project->priv->resources);
	resources = g_list_reverse (resources);
	
	return resources;
}

/* Only to be used by storage modules. */
void
mrp_project_set_resources (MrpProject *project,
			   GList      *resources)
{
	g_return_if_fail (MRP_IS_PROJECT (project));
	
	project->priv->resources = resources;

	g_list_foreach (project->priv->resources, 
			(GFunc) mpp_connect_object,
			project);
}

void      
mrp_project_add_resource (MrpProject *project, MrpResource *resource)
{
	MrpProjectPriv  *priv;
	MrpGroup        *group;
	MrpResourceType  type;
	
	g_return_if_fail (MRP_IS_PROJECT (project));
	g_return_if_fail (MRP_IS_RESOURCE (resource));
	
	priv = project->priv;

	priv->resources = g_list_prepend (priv->resources, resource);

	g_object_get (resource, "group", &group, NULL);
	
	if (!group) {
		g_object_set (resource, "group", priv->default_group, NULL);
	}

	g_object_get (resource, "type", &type, NULL);
	
	if (type == MRP_RESOURCE_TYPE_NONE) {
		g_object_set (resource, "type", MRP_RESOURCE_TYPE_WORK, NULL);
	}

	mpp_connect_object (MRP_OBJECT (resource), project);
	
	g_signal_emit (project, signals[RESOURCE_ADDED], 0, resource);

	imrp_project_set_needs_saving (project, TRUE);
}

void
mrp_project_remove_resource (MrpProject *project, MrpResource *resource)
{
	MrpProjectPriv *priv;
	
	g_return_if_fail (MRP_IS_PROJECT (project));
	g_return_if_fail (MRP_IS_RESOURCE (resource));
	
	priv = project->priv;
	
	mrp_object_removed (MRP_OBJECT (resource));

	priv->resources = g_list_remove (priv->resources, resource);

	g_signal_emit (project, signals[RESOURCE_REMOVED], 0, resource);

	imrp_project_set_needs_saving (project, TRUE);
}

GList *
mrp_project_get_groups (MrpProject *project)
{
	GList *groups;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);
	
	groups = g_list_copy (project->priv->groups);
	groups = g_list_reverse (groups);

	return groups;
}

/* Only to be used by storage modules. */
void
mrp_project_set_groups (MrpProject *project,
			GList      *groups)
{
	g_return_if_fail (MRP_IS_PROJECT (project));
	
	project->priv->groups = groups;

	g_list_foreach (project->priv->groups, 
			(GFunc) mpp_connect_object,
			project);
	
}

void
mrp_project_add_group (MrpProject *project, MrpGroup *group)
{
	MrpProjectPriv *priv;
	
	g_return_if_fail (MRP_IS_PROJECT (project));
	g_return_if_fail (MRP_IS_GROUP (group));
	
	priv = project->priv;

	priv->groups = g_list_prepend (priv->groups, group);

	g_object_set (group, "project", project, NULL);

	mpp_connect_object (MRP_OBJECT (group), project);

	g_signal_emit (project, signals[GROUP_ADDED], 0, group);

	imrp_project_set_needs_saving (project, TRUE);
}

void
mrp_project_remove_group (MrpProject *project, MrpGroup *group)
{
	MrpProjectPriv *priv;
	
	g_return_if_fail (MRP_IS_PROJECT (project));
	g_return_if_fail (MRP_IS_GROUP (group));
	
	priv = project->priv;
	
	if (priv->default_group && priv->default_group == group) {
		priv->default_group = NULL;
	}
	
	priv->groups = g_list_remove (priv->groups, group);

	g_signal_emit (project, signals[GROUP_REMOVED], 0, group);

	mrp_object_removed (MRP_OBJECT (group));

	imrp_project_set_needs_saving (project, TRUE);
}

gboolean
mrp_project_is_untouched (MrpProject *project)
{
	return project->priv->untouched;
}

gboolean
mrp_project_needs_saving (MrpProject *project)
{
	return project->priv->needs_saving;
}

const gchar *    
mrp_project_get_uri (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return project->priv->uri;
}

mrptime
mrp_project_get_project_start (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), MRP_TIME_INVALID);

	return project->priv->project_start;
}

void
mrp_project_set_project_start (MrpProject *project,
			       mrptime     start)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	project->priv->project_start = start;
}

void
imrp_project_set_needs_saving (MrpProject *project, gboolean needs_saving) 
{
	MrpProjectPriv *priv;
	
	g_return_if_fail (MRP_IS_PROJECT (project));

	priv = project->priv;

	if (needs_saving == project->priv->needs_saving) {
		return;
	}

	if (needs_saving == TRUE) {
		priv->untouched = FALSE;
	}

	project->priv->needs_saving = needs_saving;

	g_signal_emit (project, 
		       signals[NEEDS_SAVING_CHANGED], 
		       0, 
		       needs_saving);
}

/* Task related functions.
 */

GList *
mrp_project_get_all_tasks (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return mrp_task_manager_get_all_tasks (project->priv->task_manager);
}

void
mrp_project_insert_task (MrpProject *project,
			 MrpTask    *parent,
			 gint        position,
			 MrpTask    *task)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	mrp_task_manager_insert_task (project->priv->task_manager,
				      parent,
				      position,
				      task);

	g_object_set (MRP_OBJECT (task), "project", project, NULL);
}

void
mrp_project_remove_task (MrpProject *project,
			 MrpTask    *task)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	mrp_object_removed (MRP_OBJECT (task));

	mrp_task_manager_remove_task (project->priv->task_manager,
				      task);
	
	g_signal_emit (project, signals[TASK_REMOVED], 0, task);

	imrp_project_set_needs_saving (project, TRUE);
}

gboolean
mrp_project_move_task (MrpProject  *project,
		       MrpTask     *task,
		       MrpTask     *sibling,
		       MrpTask     *parent,
		       gboolean     before,
		       GError     **error)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (task), FALSE);
	g_return_val_if_fail (sibling == NULL || MRP_IS_TASK (sibling), FALSE);
	g_return_val_if_fail (MRP_IS_TASK (parent), FALSE);

	return mrp_task_manager_move_task (project->priv->task_manager,
					  task,
					  sibling,
					  parent,
					  before,
					  error);
}

void
imrp_project_task_moved (MrpProject *project,
			 MrpTask    *task)
{
	g_signal_emit (project, signals[TASK_MOVED], 0, task);

	imrp_project_set_needs_saving (project, TRUE);
}

MrpTask *
mrp_project_get_root_task (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return mrp_task_manager_get_root (project->priv->task_manager);
}

GNode *
mrp_project_get_task_tree (MrpProject *project)
{
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	return mrp_task_manager_get_task_tree (project->priv->task_manager);
}

void
mrp_project_task_traverse (MrpProject          *project,
			   MrpTask             *root,
			   MrpTaskTraverseFunc  func,
			   gpointer             user_data)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	mrp_task_manager_traverse (project->priv->task_manager,
				   root,
				   func,
				   user_data);
}

void
mrp_project_recalc_tasks (MrpProject *project)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	mrp_task_manager_recalc (project->priv->task_manager);
}

void
imrp_project_task_inserted (MrpProject *project, MrpTask *task)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	g_signal_emit (project, signals[TASK_INSERTED], 0, task);

	imrp_project_set_needs_saving (project, TRUE);
}

/* Debug function. */
void
mrp_project_dump_task_tree (MrpProject *project)
{
	g_return_if_fail (MRP_IS_PROJECT (project));

	mrp_task_manager_dump_task_tree (project->priv->task_manager);
}

MrpProperty *
imrp_project_get_property (MrpProject  *project,
			   const gchar *name,
			   GType        object_type)
{
	MrpProjectPriv *priv;
	MrpProperty    *property;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	priv = project->priv;

	property = MRP_PROPERTY (g_param_spec_pool_lookup (priv->property_pool,
							   name,
							   object_type,
							   TRUE));
	
	if (!property) {
		g_warning ("%s: object of type `%s' has no property named `%s'",
			   G_STRLOC,
			   g_type_name (object_type),
			   name);
		return NULL;
	}

	return property;
}

void
imrp_project_property_changed (MrpProject *project, MrpProperty *property)
{
	g_return_if_fail (MRP_IS_PROJECT (project));
	g_return_if_fail (property != NULL);

	g_signal_emit (project, signals[PROPERTY_CHANGED], 0, property);
	imrp_project_set_needs_saving (project, TRUE);
}

GList *
mrp_project_get_properties (MrpProject *project, GType object_type)
{
	MrpProjectPriv *priv;
	GList          *properties;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);

	priv = project->priv;

	properties = g_param_spec_pool_list_owned (priv->property_pool,
						   object_type);

	return properties;
}

void
mrp_project_add_property (MrpProject  *project,
			  GType        object_type,
			  MrpProperty *property,
                          gboolean     user_defined)
{
	MrpProjectPriv *priv;

	g_return_if_fail (MRP_IS_PROJECT (project));
	
	priv = project->priv;

	/* FIXME: Check for name conflicts. */
	if (g_param_spec_pool_lookup (priv->property_pool,
				      G_PARAM_SPEC(property)->name,
				      object_type,
				      TRUE)) {
		g_warning ("%s: Param '%s' already exists for object '%s'.",
			   G_STRLOC,
			   G_PARAM_SPEC (property)->name,
			   g_type_name (object_type));
		return;
	}

	g_param_spec_pool_insert (priv->property_pool,
				  G_PARAM_SPEC (property),
				  object_type);

	imrp_property_set_project (property, project);

	g_signal_emit (project, signals[PROPERTY_ADDED], 0, property);

        if (user_defined) {
                imrp_project_set_needs_saving (project, TRUE);
        }
}

void
mrp_project_remove_property (MrpProject  *project,
			     GType        object_type,
			     const gchar *name)
{
	MrpProjectPriv *priv;
	MrpProperty    *property;
	
	g_return_if_fail (MRP_IS_PROJECT (project));
	
	priv = project->priv;

	property = imrp_project_get_property (project, name, object_type);

	if (!property) {
		g_warning ("%s: object of type '%s' has no property named '%s'",
			   G_STRLOC,
			   g_type_name (object_type),
			   name);
		return;
	}

	g_signal_emit (project, signals[PROPERTY_REMOVED], 0, property);
	
	g_param_spec_pool_remove (priv->property_pool,
				  G_PARAM_SPEC (property));

	imrp_project_set_needs_saving (project, TRUE);
}
	
MrpCalendar *
mrp_project_get_calendar (MrpProject *project)
{
	MrpProjectPriv *priv;
	
	g_return_val_if_fail (MRP_IS_PROJECT (project), NULL);
	
	priv = project->priv;
	
	return priv->calendar;
}


