/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2001-2002  CodeFactory AB
 * Copyright (C) 2001-2002  Richard Hult <rhult@codefactory.se>
 * Copyright (C) 2001-2002  Mikael Hallendal <micke@codefactory.se>
 * Copyright (C) 2002       Alvaro del Castillo <acs@barrapunto.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#include <config.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <glib.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlmemory.h>
#include "mrproject/mrp-private.h"
#include "mrproject/mrp-error.h"
#include "mrproject/mrp-intl.h"
#include "mrproject/mrp-task.h"
#include "mrproject/mrp-resource.h"
#include "mrproject/mrp-group.h"
#include "mrproject/mrp-relation.h"
#include "mrp-parser.h"

/* WARNING: This code is a hack just to have a file loader/saver for the old
 * format. Don't expect to understand any of the code or anything. It sucks.
 */

typedef struct {
        xmlDocPtr       doc;

	gint            version;
	
	MrpProject     *project;
	
	MrpTask        *root_task;
	GList          *resources;
	GList          *groups;
	GList          *assignments;

	mrptime         project_start;
	
	MrpGroup       *default_group;

	gint            last_id;
	
	GHashTable     *task_hash;
	GHashTable     *resource_hash;
	GHashTable     *group_hash;
	GList          *delayed_relations;
} MrpParser;

typedef struct {
	xmlNodePtr node;
	gint       id;
} NodeEntry;


static gchar           *mpp_xml_get_string            (xmlNodePtr   node,
						       const char  *name);
static gint             mpp_xml_get_int               (xmlNodePtr   node,
						       const char  *name);
static gint             mpp_xml_get_int_with_default  (xmlNodePtr   node,
						       const char  *name,
						       gint         def);
static mrptime          mpp_xml_get_date              (xmlNodePtr   node,
						       const char  *name);
static gfloat           mpp_xml_get_float             (xmlNodePtr   node,
						       const char  *name);
static void             mpp_xml_set_date              (xmlNodePtr   node,
						       const gchar *prop,
						       mrptime      time);
static void             mpp_xml_set_int               (xmlNodePtr   node,
						       const gchar *prop,
						       gint         value);
static void             mpp_xml_set_float             (xmlNodePtr   node,
						       const gchar *prop, 
						       gfloat       value);
static xmlNodePtr       mpp_xml_search_child          (xmlNodePtr   node, 
						       const gchar *name);
static MrpPropertyType  mpp_property_type_from_string (const gchar *str);
static gchar           *mpp_property_to_string        (MrpObject   *object,
						       MrpProperty *property);
static void             mpp_set_property_from_string  (MrpProject  *project,
						       MrpObject   *object,
						       const gchar *name,
						       const gchar *value);


static void
mpp_read_predecessor (MrpParser  *parser,
		      gint        task_id,
		      xmlNodePtr  tree)
{
	gint             predecessor_id;
	DelayedRelation *relation;
	gchar           *str;
	MrpRelationType  type;

	if (strcmp (tree->name, "predecessor")){
		/*g_warning ("Got %s, expected 'predecessor'.", tree->name);*/
		return;
	}

	predecessor_id = mpp_xml_get_int (tree, "predecessor-id");
	if (predecessor_id == 0) {
		g_warning ("Invalid predecessor read.");
		return;
	}

	str = mpp_xml_get_string (tree, "type");
	if (!strcmp (str, "FS")) {
		type = MRP_RELATION_FS;
	}
	else if (!strcmp (str, "FF")) {
		type = MRP_RELATION_FF;
	}
	else if (!strcmp (str, "SS")) {
		type = MRP_RELATION_SS;
	}
	else if (!strcmp (str, "SF")) {
		type = MRP_RELATION_SF;
	}
	else {
		g_warning ("Invalid dependency type.");
		g_free (str);
		return;
	}

	g_free (str);

	relation = g_new0 (DelayedRelation, 1);
	relation->successor_id = task_id;
	relation->predecessor_id = predecessor_id;
	relation->type = type;
	
	parser->delayed_relations = g_list_prepend (parser->delayed_relations,
						    relation);
}

static gboolean
mpp_read_constraint (xmlNodePtr node, MrpConstraint *constraint)
{
	gchar *str;
	
	str = mpp_xml_get_string (node, "type");

	if (str == NULL) {
		g_warning ("Invalid constraint read.");
		return FALSE;
	}

	if (!strcmp (str, "must-start-on")) {
		constraint->type = MRP_CONSTRAINT_MSO;
	} else if (!strcmp (str, "start-no-earlier-than")) {
		constraint->type = MRP_CONSTRAINT_SNET;
	} else if (!strcmp (str, "finish-no-later-than")) {
		constraint->type = MRP_CONSTRAINT_FNLT;
	} else {
		g_warning ("Cant handle constraint '%s'", str);
		g_free (str);

		return FALSE;
	}

	constraint->time = mpp_xml_get_date (node, "time");

	return TRUE;
}

static void
mpp_read_custom_properties (MrpParser  *parser,
			    xmlNodePtr  node,
			    MrpObject  *object)
{
	xmlNodePtr  child;
	gchar      *name;
	gchar      *value;

	node = mpp_xml_search_child (node, "properties"); 

	if (node == NULL) {
		return;
	}
	
	for (child = node->children; child; child = child->next) {
		if (strcmp (child->name, "property")) {
			continue;
		}
		
		name  = mpp_xml_get_string (child, "name");
		value = mpp_xml_get_string (child, "value");

		mpp_set_property_from_string (parser->project,
					      object,
					      name,
					      value);
		
		g_free (name);
		g_free (value);
	}
}

static void
mpp_read_task (MrpParser *parser, xmlNodePtr tree, MrpTask *parent)
{
	xmlNodePtr     tasks, predecessor;
	gchar          *name;
	gint           id;
	mrptime        start, end;
	MrpTask       *task;
	MrpConstraint  constraint;
	guint          percent_complete = 0;
	gchar         *note;
	gint           duration, work;
	gboolean       got_constraint = FALSE;
	
	if (strcmp (tree->name, "task")){
		/*g_warning ("Got %s, expected 'task'.", tree->name);*/
		return;
	}

	name = mpp_xml_get_string (tree, "name");
	note = mpp_xml_get_string (tree, "note");
	id = mpp_xml_get_int (tree, "id");
	percent_complete = mpp_xml_get_int (tree, "percent-complete");

	if (parser->version == 1) {
		start = mpp_xml_get_date (tree, "start");
		end = mpp_xml_get_date (tree, "end");
		duration = end - start;

		if (parser->project_start == -1) {
			parser->project_start = start;
		} else {
			parser->project_start = MIN (parser->project_start, start);
		}

		constraint.type = MRP_CONSTRAINT_MSO;
		constraint.time = start;
		got_constraint = TRUE;
		
		task = g_object_new (MRP_TYPE_TASK,
				     "project", parser->project,
				     "name", name,
				     "duration", duration,
				     "percent-complete", percent_complete,
				     "note", note,
				     NULL);
	} else {
		/* Use work if available, otherwise use duration. */
		work = mpp_xml_get_int_with_default (tree, "work", -1);
		duration = 0;

		if (work == -1) {
			duration = mpp_xml_get_int (tree, "duration");
		}
		
		task = g_object_new (MRP_TYPE_TASK,
				     "project", parser->project,
				     "name", name,
				     /* A bit ugly but hey... :) */
				     work != -1 ? "work" : "duration",
				     work != -1 ? work : duration,
				     "percent-complete", percent_complete,
				     "note", note,
				     NULL);
	}
	
	g_free (name);
	g_free (note);

	/* Note: We should use mrp_project_insert_task here instead of the
	 * private imrp_task_insert_child, and not set the "project" property
	 * above. But then the project will emit signals to the views etc,
	 * maybe we should have freeze/thaw on the project or something.
	 */
	imrp_task_insert_child (parent, -1, task);

	/* Treat duration from old files as work. */
	if (parser->version == 1) {
		gint work;

		work = mrp_task_duration_to_work (task, duration);
		
		g_object_set (G_OBJECT (task),
			      "work", work,
			      NULL);
	}
	
	g_hash_table_insert (parser->task_hash, GINT_TO_POINTER (id), task);

	mpp_read_custom_properties (parser, tree, MRP_OBJECT (task));

	for (tasks = tree->children; tasks; tasks = tasks->next) {
		if (!strcmp (tasks->name, "task")) {
			mpp_read_task (parser, tasks, task);
		}
		else if (!strcmp (tasks->name, "predecessors")) {
			for (predecessor = tasks->children; predecessor; predecessor = predecessor->next) {
				mpp_read_predecessor (parser, id, predecessor);
			}
		}
		else if (!strcmp (tasks->name, "constraint")) {
			got_constraint = mpp_read_constraint (tasks, &constraint);
		}
	}

	if (got_constraint) {
		g_object_set (G_OBJECT (task),
			      "constraint", &constraint,
			      NULL);
	}
}

static void
mpp_read_resource (MrpParser *parser, xmlNodePtr tree)
{
	gint         id;
	gint         type;
	gchar       *name, *email;
	gint         gid;
	gint         units;
	gfloat       std_rate; /*, ovt_rate;*/
	MrpResource *resource;
	MrpGroup    *group;

	if (strcmp (tree->name, "resource")){
		/*g_warning ("Got %s, expected 'resource'.", tree->name);*/
		return;
	}

	id       = mpp_xml_get_int (tree, "id");
	name     = mpp_xml_get_string (tree, "name");
	gid      = mpp_xml_get_int (tree, "group");
	type     = mpp_xml_get_int (tree, "type");
	units    = mpp_xml_get_int (tree, "units");
	std_rate = mpp_xml_get_float (tree, "std-rate");
	/*ovt_rate = mpp_xml_get_float (tree, "ovt-rate");*/
	email    = mpp_xml_get_string (tree, "email");

	if (email == NULL) {
		email = g_strdup ("");
	}

	group = g_hash_table_lookup (parser->group_hash, GINT_TO_POINTER (gid));

	resource = g_object_new (MRP_TYPE_RESOURCE,
				 "name", name,
				 "type", type,
				 "group", group,
				 "units", units,
				 "email", email,
				 NULL);
	
	/* These are cost custom properties */
	/* We need first to register the resource in the project */
	mrp_project_add_resource (parser->project, resource);
	
	mrp_object_set (MRP_OBJECT (resource),
			"cost", std_rate,
			/*"cost_overtime", ovt_rate,*/
			NULL);

	g_hash_table_insert (parser->resource_hash, 
			     GINT_TO_POINTER (id), resource);

	parser->resources = g_list_prepend (parser->resources, resource);
	
	g_free (name);
	g_free (email);
}

static void
mpp_read_group (MrpParser *parser, xmlNodePtr tree)
{
	gint      id;
	gchar    *name;
	gchar    *mgr_name, *mgr_phone, *mgr_email;
	MrpGroup *group;
        
	if (strcmp (tree->name, "group")){
		/*g_warning ("Got %s, expected 'group'.", tree->name);*/
		return;
	}

	id = mpp_xml_get_int (tree, "id");

	name = mpp_xml_get_string (tree, "name");
	mgr_name  = mpp_xml_get_string (tree, "admin-name");
	mgr_phone = mpp_xml_get_string (tree, "admin-phone");
	mgr_email = mpp_xml_get_string (tree, "admin-email");

	group = g_object_new (MRP_TYPE_GROUP,
			      "name", name,
			      "manager_name", mgr_name,
			      "manager_phone", mgr_phone,
			      "manager_email", mgr_email,
			      NULL);

	g_free (name);
	g_free (mgr_name);
	g_free (mgr_phone);
	g_free (mgr_email);
	
	g_hash_table_insert (parser->group_hash, GINT_TO_POINTER (id), group);

	parser->groups = g_list_prepend (parser->groups, group);
}

static void
mpp_read_assignment (MrpParser *parser, xmlNodePtr tree)
{
	gint   task_id, resource_id;
	MrpAssignment *assignment;
	MrpTask       *task;
	MrpResource   *resource;

	if (strcmp (tree->name, "allocation")){
		/*g_warning ("Got %s, expected 'allocation'.", tree->name);*/
		return;
	}

	task_id = mpp_xml_get_int (tree, "task-id");
	resource_id = mpp_xml_get_int (tree, "resource-id");
	
	task = g_hash_table_lookup (parser->task_hash, 
				    GINT_TO_POINTER (task_id));
	resource = g_hash_table_lookup (parser->resource_hash,
					GINT_TO_POINTER (resource_id));
	
	if (!task) {
		g_print ("NO TASK FROM HASH\n");
	}

	if (!resource) {
		g_print ("NO RESOURCE FROM HASH\n");
	}

	assignment = g_object_new (MRP_TYPE_ASSIGNMENT,
				   "task", task,
				   "resource", resource,
				   NULL);

	parser->assignments = g_list_prepend (parser->assignments, assignment);
}

static void
mpp_read_project_properties (MrpParser *parser)
{
	xmlNodePtr  node;
	gchar      *name;
	gchar      *org;
	gchar      *manager;

	node = parser->doc->children;

	parser->version = mpp_xml_get_int_with_default (node,
							"mrproject-version",
							1);

	if (parser->version > 1) {
		parser->project_start = mpp_xml_get_date (node,
							  "project-start");
	}
	
	name = mpp_xml_get_string (node, "name");
	org = mpp_xml_get_string (node, "company");
	manager = mpp_xml_get_string (node, "manager");

	g_object_set (G_OBJECT (parser->project),
		      "name", name,
		      "organization", org,
		      "manager", manager,
		      NULL);

	g_free (name);
	g_free (org);
	g_free (manager);
}

static void
mpp_read_property_specs (MrpParser *parser)
{
	xmlNodePtr       node;
	xmlNodePtr       child;
	gchar           *name;
	gchar           *label;
	gchar           *description;
	gchar           *owner_str;
	gchar           *type_str;
	MrpProperty     *property;
	MrpPropertyType  type;
		
	node = mpp_xml_search_child (parser->doc->children, "properties"); 

	if (node == NULL) {
		return;
	}
	
	for (child = node->children; child; child = child->next) {
		if (strcmp (child->name, "property")) {
			continue;
		}
		
		name        = mpp_xml_get_string (child, "name");
		label       = mpp_xml_get_string (child, "label");
		description = mpp_xml_get_string (child, "description");
		owner_str   = mpp_xml_get_string (child, "owner");
		type_str    = mpp_xml_get_string (child, "type");

		type = mpp_property_type_from_string (type_str);
		
		property = mrp_property_new (name,
					     type,
					     label,
					     description);
		     
		/* FIXME: support project and resource props as well. */
		mrp_project_add_property (parser->project,
					  MRP_TYPE_TASK, 
					  property,
					  FALSE /* FIXME: user_defined, should 
						   be read from the file */);

		g_free (name);
		g_free (type_str);
		g_free (owner_str);
		g_free (label);
		g_free (description);
	}
}

static gboolean
mpp_read_project (MrpParser *parser)
{
	xmlNodePtr  child, tasks, resources, groups, assignments;
	gint        gid;
	
	mpp_read_project_properties (parser);

	parser->root_task = mrp_task_new ();

	mpp_read_property_specs (parser);

	/* Load tasks. */
	child = mpp_xml_search_child (parser->doc->children, "tasks");
	if (child != NULL) {
		for (tasks = child->children; tasks; tasks = tasks->next) {
			mpp_read_task (parser, tasks, parser->root_task);
		}
	}

	/* Note: if we read a G1 file, there was no project-start set, so make
	 * sure the one we faked from the earliest task starts on at 00:00.
	 */
	if (parser->version == 1) {
		mrp_time_align_day (parser->project_start);
	}
	
	/* Load resource groups. */
 	child = mpp_xml_search_child (parser->doc->children, 
				      "resource-groups"); 
 	if (child != NULL) {
		for (groups = child->children; groups; groups = groups->next) {
			mpp_read_group (parser, groups); 
		}
	}

	gid = mpp_xml_get_int (child, "default_group");
        parser->default_group = g_hash_table_lookup (parser->group_hash,
						     GINT_TO_POINTER (gid));

	/* Load resources. */
 	child = mpp_xml_search_child (parser->doc->children, "resources"); 
	if (child != NULL) {
		for (resources = child->children; resources; resources = resources->next) {
			mpp_read_resource (parser, resources); 
		}
	}
	parser->resources = g_list_reverse (parser->resources);
	
	/* Load assignments. */
	child = mpp_xml_search_child (parser->doc->children, "allocations");
	if (child != NULL) {
		for (assignments = child->children; assignments; assignments = assignments->next) {
			mpp_read_assignment (parser, assignments);
		}
	}

	return TRUE;
}

gboolean
mrp_parser_load (MrpStorageMrproject *module, const gchar *uri, GError **error)
{
	MrpParser    parser;
	const gchar *filename;
	gboolean     success;
	
	g_return_val_if_fail (MRP_IS_STORAGE_MRPROJECT (module), FALSE);
	g_return_val_if_fail (uri != NULL, FALSE);

	if (strncmp (uri, "mrproject-1:", 12) != 0) {
		g_warning ("Mrproject-1 format parser can only handle local files (%s).", uri);
		return FALSE;
	}

	memset (&parser, 0, sizeof (parser));

	parser.project_start = -1;
	parser.project = module->project;
	
	filename = uri + 12; /* Skip mrproject-1: */
	if (strncmp (filename, "///", 3) == 0) {
		/* Skip extra // */
		filename += 2;
	}
	
	parser.doc = xmlParseFile (filename);
	
	if (parser.doc == NULL) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_LOAD_FILE_DONT_EXIST,
			     _("Failed to load file %s."), filename);
		return FALSE;
	}

	if (parser.doc->children == NULL) {
		xmlFreeDoc (parser.doc);
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_LOAD_FILE_INVALID,
			     _("%s is not valid XML."), 
			     filename);
		return FALSE;
	}

	parser.task_hash = g_hash_table_new (NULL, NULL);
	parser.resource_hash = g_hash_table_new (NULL, NULL);
	parser.group_hash = g_hash_table_new (NULL, NULL);
	
	success = mpp_read_project (&parser);

	g_hash_table_destroy (parser.resource_hash);
	g_hash_table_destroy (parser.group_hash);

	xmlFreeDoc (parser.doc);

	if (!success) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_FAILED,
			     _("\"%s\" is not a valid MrProject file"),
			     filename);
		return FALSE;
	}
	
	module->project_start = parser.project_start;
	module->root_task = parser.root_task;
	module->task_id_hash = parser.task_hash;
	module->delayed_relations = parser.delayed_relations;
	module->groups = parser.groups;
	module->resources = parser.resources;
	module->assignments = parser.assignments;
	module->default_group = parser.default_group;

	g_object_set_data (G_OBJECT (parser.project),
			   "version",
			   GINT_TO_POINTER (parser.version));
	
	return TRUE;
}

/****************
 * Saving.
 */

static void
mpp_write_project_properties (MrpParser *parser, xmlNodePtr node)
{
	gchar   *name;
	gchar   *org;
	gchar   *manager;
	mrptime  pstart;

	g_object_get (parser->project,
		      "name", &name,
		      "organization", &org,
		      "manager", &manager,
		      "project-start", &pstart,
		      NULL);

	xmlSetProp (node, "name", name);
	xmlSetProp (node, "company", org);
	xmlSetProp (node, "manager", manager);

	mpp_xml_set_date (node, "project-start", pstart);
	mpp_xml_set_int (node, "mrproject-version", 2);

	g_free (name);
	g_free (org);
	g_free (manager);
}

static const gchar *
mpp_property_type_to_string (MrpPropertyType type)
{
	switch (type) {
	case MRP_PROPERTY_TYPE_INT:
		return "int";
	case MRP_PROPERTY_TYPE_STRING:
		return "text";
	case MRP_PROPERTY_TYPE_FLOAT:
		return "float";
	case MRP_PROPERTY_TYPE_DURATION:
		return "duration";
	case MRP_PROPERTY_TYPE_DATE:
		return "date";	
	default:
		g_warning ("Not implemented support for type");
		break;
	}

	return NULL;
}

static MrpPropertyType
mpp_property_type_from_string (const gchar *str)
{
	if (!strcmp (str, "int")) {
		return MRP_PROPERTY_TYPE_INT;
	}
	else if (!strcmp (str, "text")) {
		return MRP_PROPERTY_TYPE_STRING;
	}
	else if (!strcmp (str, "float")) {
		return MRP_PROPERTY_TYPE_FLOAT;
		}
	else if (!strcmp (str, "date")) {
		return MRP_PROPERTY_TYPE_DATE;
			}
	else if (!strcmp (str, "duration")) {
		return MRP_PROPERTY_TYPE_DURATION;
	} else {	
		g_warning ("Not implemented support for type");
	}

	return MRP_PROPERTY_TYPE_NONE;
}

static gchar *
mpp_property_to_string (MrpObject   *object,
			MrpProperty *property)
{
	const gchar *name;
	gchar       *str;
	gchar        buffer[G_ASCII_DTOSTR_BUF_SIZE];
	gint         i;
	gfloat       f;
	mrptime      date;
	
	name = mrp_property_get_name (property);
	
	switch (mrp_property_get_type (property)) {
	case MRP_PROPERTY_TYPE_STRING:
		mrp_object_get (object, name, &str, NULL);
		return str;
	case MRP_PROPERTY_TYPE_INT:
		mrp_object_get (object, name, &i, NULL);
		return g_strdup_printf ("%d", i);
	case MRP_PROPERTY_TYPE_FLOAT:
		mrp_object_get (object, name, &f, NULL);
		g_ascii_dtostr (buffer, sizeof (buffer), (double) f);
		return g_strdup (buffer);
	case MRP_PROPERTY_TYPE_DURATION:
		mrp_object_get (object, name, &i, NULL);
		return g_strdup_printf ("%d", i);
	case MRP_PROPERTY_TYPE_DATE:
		mrp_object_get (object, name, &date, NULL);
		str = mrp_time_to_string (date);		
		return str;
	default:
		g_warning ("Not implemented support for type.");
		break;
	}

	return NULL;
}

static void
mpp_set_property_from_string (MrpProject  *project,
			      MrpObject   *object,
			      const gchar *name,
			      const gchar *value)
{
	MrpProperty     *property;
	MrpPropertyType  type;
	mrptime          date;
	gint             i;
	gfloat           f;

	property = imrp_project_get_property (project,
					      name,
					      G_OBJECT_TYPE (object));

	if (property == NULL) {
		return;
	}
	
	type = mrp_property_get_type (property);
	
	switch (type) {
	case MRP_PROPERTY_TYPE_STRING:
		mrp_object_set (object, name, value, NULL);
		break;
	case MRP_PROPERTY_TYPE_INT:
		i = atoi (value);
		mrp_object_set (object, name, i, NULL);
		break;
	case MRP_PROPERTY_TYPE_FLOAT:
		f = g_ascii_strtod (value, NULL);
		mrp_object_set (object, name, f, NULL);
		break;
	case MRP_PROPERTY_TYPE_DURATION:
		i = atoi (value);
		mrp_object_set (object, name, i, NULL);
		break;
	case MRP_PROPERTY_TYPE_DATE:
		date = mrp_time_from_string (value, NULL);
		mrp_object_set (object, name, &date, NULL);
		break;
	default:
		g_warning ("Not implemented support for type.");
		break;
	}
}

static void
mpp_write_property_specs (MrpParser *parser, xmlNodePtr node)
{
	GList           *properties, *l;
	xmlNodePtr       child;
	MrpProperty     *property;
	MrpPropertyType  type;
	
	/* FIXME: support project and resource props as well. */
	properties = mrp_project_get_properties (parser->project,
						 MRP_TYPE_TASK);

	if (properties == NULL) {
		return;
	}
	
	node = xmlNewChild (node, NULL, "properties", NULL);

	for (l = properties; l; l = l->next) {
		property = l->data;
		
		child = xmlNewChild (node, NULL, "property", NULL);
		
		xmlSetProp (child, "name", mrp_property_get_name (property));

		type = mrp_property_get_type (property);
		xmlSetProp (child, "type", mpp_property_type_to_string (type));
				    
		xmlSetProp (child, "owner", "task");
		xmlSetProp (child, "label", mrp_property_get_label (property));
		xmlSetProp (child, "description", mrp_property_get_description (property));
	}

	g_list_free (properties);
}

static void
mpp_write_predecessor (MrpParser   *parser,
		       xmlNodePtr   node,
		       MrpRelation *relation)
{
	gchar     *str;
	NodeEntry *entry;
	gint       lag;

	node = xmlNewChild (node, NULL, "predecessor", NULL);

	xmlSetProp (node, "id", "1"); /* Don't need id here. */

	entry = g_hash_table_lookup (parser->task_hash,
				     mrp_relation_get_predecessor (relation));
	mpp_xml_set_int (node, "predecessor-id", entry->id);

	switch (mrp_relation_get_relation_type (relation)) {
	case MRP_RELATION_FS:
		str = "FS";
		break;
	case MRP_RELATION_FF:
		str = "FF";
		break;
	case MRP_RELATION_SF:
		str = "SF";
		break;
	case MRP_RELATION_SS:
		str = "SS";
		break;
	default:
		str = "FS";
	}

	xmlSetProp (node, "type", str);

	lag = mrp_relation_get_lag (relation);
	if (lag) {
		mpp_xml_set_int (node, "lag", lag);
	}
}

static gboolean
mpp_hash_insert_task_cb (MrpTask *task, MrpParser *parser)
{
	NodeEntry *entry;
	
	/* Don't want the root task. */
	if (task == parser->root_task) {
		return FALSE;
	}
	
	entry = g_new0 (NodeEntry, 1);
	entry->id = parser->last_id++;
	
	g_hash_table_insert (parser->task_hash, task, entry);
	
	return FALSE;
}

static void
mpp_write_constraint (xmlNodePtr node, MrpConstraint *constraint)
{
	xmlNodePtr   child;
	const gchar *str = NULL;
	
	/* No need to save if we have ASAP. */
	if (constraint->type == MRP_CONSTRAINT_ASAP) {
		return;
	}

	child = xmlNewChild (node, NULL, "constraint", NULL);

	switch (constraint->type) {
	case MRP_CONSTRAINT_MSO:
		str = "must-start-on";
		break;
	case MRP_CONSTRAINT_SNET:
		str = "start-no-earlier-than";
		break;
	case MRP_CONSTRAINT_FNLT:
		str = "finish-no-later-than";
		break;
	case MRP_CONSTRAINT_ASAP:
	case MRP_CONSTRAINT_ALAP:
		g_assert_not_reached ();
		break;
	}	

	xmlSetProp (child, "type", str);
	mpp_xml_set_date (child, "time", constraint->time);
}

static void
mpp_write_custom_properties (MrpParser  *parser,
			     xmlNodePtr  node,
			     MrpObject  *object)
{
	GList           *properties, *l;
	xmlNodePtr       child;
	MrpProperty     *property;
	gchar           *value;

	properties = mrp_project_get_properties (parser->project,
						 G_OBJECT_TYPE (object));
	
	if (properties == NULL) {
		return;
	}
	
	node = xmlNewChild (node, NULL, "properties", NULL);

	for (l = properties; l; l = l->next) {
		property = l->data;
		
		child = xmlNewChild (node, NULL, "property", NULL);
		
		xmlSetProp (child, "name", mrp_property_get_name (property));

		value = mpp_property_to_string (object, property);
		
		xmlSetProp (child, "value", value);

		g_free (value);
	}

	g_list_free (properties);
}

static gboolean
mpp_write_task_cb (MrpTask *task, MrpParser *parser)
{
	MrpTask       *parent;
	NodeEntry     *entry;
	xmlNodePtr     node, parent_node;
	gchar         *name;
	gchar         *note;
	mrptime        start, finish;
	MrpConstraint *constraint;
	gint           duration;
	gint           work;
	gint           complete;
	GList         *predecessors, *l;

	/* Don't want the root task. */
	if (task == parser->root_task) {
		return FALSE;
	}
	
	parent = mrp_task_get_parent (task);

	entry = g_hash_table_lookup (parser->task_hash, parent);
	parent_node = entry->node;

	node = xmlNewChild (parent_node, NULL, "task", NULL);

	entry = g_hash_table_lookup (parser->task_hash, task);
	entry->node = node;
	
	g_object_get (task,
		      "name", &name,
		      "note", &note,
		      "start", &start,
		      "duration", &duration,
		      "work", &work,
		      "constraint", &constraint,
		      "percent-complete", &complete,
		      NULL);

	finish = start + duration;
	
	mpp_xml_set_int (node, "id", entry->id);
	xmlSetProp (node, "name", name);
	xmlSetProp (node, "note", note);
	mpp_xml_set_int (node, "work", work);

	/* These are set for G1 version, they are not used when loading a G2
	 * file.
	 */
	mpp_xml_set_date (node, "start", start);
	mpp_xml_set_date (node, "end", finish);

	mpp_xml_set_int (node, "percent-complete", complete);

	mpp_write_custom_properties (parser, node, MRP_OBJECT (task));
	
	mpp_write_constraint (node, constraint);
	
	predecessors = mrp_task_get_predecessors (task);
	if (predecessors != NULL) {
		node = xmlNewChild (node, NULL, "predecessors", NULL);
		for (l = predecessors; l; l = l->next) {
			mpp_write_predecessor (parser, node, l->data);
		}
	}

	g_list_free (predecessors);
	
	return FALSE;
}

static void
mpp_hash_insert_group (MrpParser *parser, MrpGroup *group)
{
	NodeEntry *entry;

	entry = g_new0 (NodeEntry, 1);

	entry->id = parser->last_id++;

	g_hash_table_insert (parser->group_hash, group, entry);
}

static void
mpp_write_group (MrpParser *parser, xmlNodePtr parent, MrpGroup *group)
{
	NodeEntry  *entry;
	xmlNodePtr  node;
	gchar      *name, *admin_name, *admin_phone, *admin_email;

	g_return_if_fail (MRP_IS_GROUP (group));

	node = xmlNewChild (parent,
                            NULL,
			    "group",
			    NULL);

	entry = g_hash_table_lookup (parser->group_hash, group);
	entry->node = node;
	
	mpp_xml_set_int (node, "id", entry->id);
	
	g_object_get (G_OBJECT (group),
		      "name", &name,
		      "manager-name", &admin_name,
		      "manager-phone", &admin_phone,
		      "manager-email", &admin_email,
		      NULL);

	xmlSetProp (node, "name", name);	
	xmlSetProp (node, "admin-name", admin_name);	
	xmlSetProp (node, "admin-phone", admin_phone);	
	xmlSetProp (node, "admin-email", admin_email);	
}

static void
mpp_hash_insert_resource (MrpParser *parser, MrpResource *resource)
{
	NodeEntry  *entry;

	entry = g_new0 (NodeEntry, 1);
	entry->id = parser->last_id++;

	g_hash_table_insert (parser->resource_hash, resource, entry);
}

static void
mpp_write_resource (MrpParser   *parser, 
		    xmlNodePtr   parent,
		    MrpResource *resource)
{
	xmlNodePtr  node;
	gchar      *name, *email;
	gint        type, units;
	gfloat      std_rate; /*, ovt_rate;*/
	NodeEntry  *group_entry;
	NodeEntry  *resource_entry;
	MrpGroup   *group;

	g_return_if_fail (MRP_IS_RESOURCE (resource));

	node = xmlNewChild (parent,
                            NULL,
			    "resource",
			    NULL);

	mrp_object_get (MRP_OBJECT (resource),
			"name", &name,
			"email", &email,
			"type", &type,
			"units", &units,
			"group", &group,
			"cost", &std_rate,
			/*"cost-overtime", &ovt_rate,*/
			NULL);

	group_entry = g_hash_table_lookup (parser->group_hash, group);

	/* FIXME: should group really be able to be NULL? Should always 
	 * be default group? */
	if (group_entry != NULL) {
		mpp_xml_set_int (node, "group", group_entry->id);
	}

	resource_entry = g_hash_table_lookup (parser->resource_hash, resource);
	mpp_xml_set_int (node, "id", resource_entry->id);
	
	xmlSetProp (node, "name", name);
	
	mpp_xml_set_int (node, "type", type);
	
	mpp_xml_set_int (node, "units", units);
	xmlSetProp (node, "email", email);

	mpp_xml_set_float (node, "std-rate", std_rate);
	/*mpp_xml_set_float (node, "ovt-rate", ovt_rate);*/
}

static void
mpp_write_assignment (MrpParser     *parser,
		      xmlNodePtr     parent,
		      MrpAssignment *assignment)
{
	xmlNodePtr   node;
	MrpTask     *task;
	MrpResource *resource;
	NodeEntry   *resource_entry;
	NodeEntry   *task_entry;

	g_return_if_fail (MRP_IS_ASSIGNMENT (assignment));

	node = xmlNewChild (parent,
                            NULL,
			    "allocation",
			    NULL);

	g_object_get (G_OBJECT (assignment),
		      "task", &task,
		      "resource", &resource,
		      NULL);

	task_entry = g_hash_table_lookup (parser->task_hash, task);
	resource_entry = g_hash_table_lookup (parser->resource_hash, resource);

	mpp_xml_set_int (node, "task-id", task_entry->id);
	mpp_xml_set_int (node, "resource-id", resource_entry->id);
}

static gboolean
mpp_write_project (MrpParser *parser)
{
	xmlNodePtr  node, child;
	GList      *list, *l;
	GList      *assignments = NULL;
	MrpGroup   *default_group = NULL;
	NodeEntry  *entry;
	
	node = xmlNewDocNode (parser->doc, NULL, "project", NULL);
	parser->doc->xmlRootNode = node;

	mpp_write_project_properties (parser, node);

	mpp_write_property_specs (parser, node);
	
	/* Write tasks. */

 	child = xmlNewChild (node, NULL, "tasks",NULL);

	entry = g_new0 (NodeEntry, 1);
	entry->id = 0;
	entry->node = child;

	g_hash_table_insert (parser->task_hash, parser->root_task, entry);
	
	/* Generate IDs and hash table. */
	parser->last_id = 1;
	mrp_project_task_traverse (parser->project,
				   parser->root_task,
				   (MrpTaskTraverseFunc) mpp_hash_insert_task_cb,
				   parser);
	
	mrp_project_task_traverse (parser->project,
				   parser->root_task,
				   (MrpTaskTraverseFunc) mpp_write_task_cb,
				   parser);

	/* Write resource groups. */

 	child = xmlNewChild (node, NULL, "resource-groups",NULL);
	list = mrp_project_get_groups (parser->project);

	/* Generate IDs and hash table. */
	parser->last_id = 1;
	for (l = list; l; l = l->next) {
		mpp_hash_insert_group (parser, l->data); 
	}

	g_object_get (parser->project, "default-group", &default_group, NULL);

	if (default_group) {
		entry = g_hash_table_lookup (parser->group_hash, 
					     default_group);
		mpp_xml_set_int (child, "default_group", entry->id);
	}
	
	for (l = list; l; l = l->next) {
		mpp_write_group (parser, child, l->data); 
	}
	g_list_free (list);
	
	/* Write resources. */

 	child = xmlNewChild (node, NULL, "resources",NULL);
	list = mrp_project_get_resources (parser->project);

	/* Generate IDs and hash table. */
	parser->last_id = 1;
	for (l = list; l; l = l->next) {
		GList *r_list;
		mpp_hash_insert_resource (parser, l->data); 
		r_list = mrp_resource_get_assignments (MRP_RESOURCE (l->data));
		assignments = g_list_concat (assignments, r_list);
	}

	for (l = list; l; l = l->next) {
		mpp_write_resource (parser, child, l->data); 
	}
	g_list_free (list);

	/* Write assignments. */
 	child = xmlNewChild (node, NULL, "allocations",NULL);

	for (l = assignments; l; l = l->next) {
		mpp_write_assignment (parser, child, l->data);
	}

	g_list_free (assignments);

	return TRUE;
}

gboolean
mrp_parser_save (MrpStorageMrproject  *module,
		 const gchar          *uri,
		 gboolean              force,
		 GError              **error)
{
	MrpParser    parser;
	const gchar *filename;
	gchar       *real_filename;
	gint         ret;
	gboolean     file_exist;
	
	g_return_val_if_fail (MRP_IS_STORAGE_MRPROJECT (module), FALSE);
	g_return_val_if_fail (uri != NULL, FALSE);

	if (strncmp (uri, "mrproject-1:", 12) != 0) {
		g_warning ("Mrproject-1 parser can only handle local files (%s).", uri);
		return FALSE;
	}
	
	/* We want indentation. */
	xmlKeepBlanksDefault (0);
	
	memset (&parser, 0, sizeof (parser));

	parser.project = module->project;
	parser.task_hash = g_hash_table_new_full (NULL, NULL, NULL, g_free);
	parser.group_hash = g_hash_table_new_full (NULL, NULL, NULL, g_free);
	parser.resource_hash = g_hash_table_new_full (NULL, NULL, NULL, 
						      g_free);
	parser.root_task = mrp_project_get_root_task (parser.project);

	filename = uri + 12; /* Skip mrproject-1: */

	if (strncmp (filename, "///", 3) == 0) {
		/* Skip extra // */
		filename += 2;
	}

	if (!strstr (filename, ".mrproject")) {
		real_filename = g_strconcat (filename, ".mrproject", NULL);
	} else {
		real_filename = g_strdup (filename);
	}
	
	/* FIXME: Check if file exist, if it does, return an error message */
	file_exist = g_file_test (real_filename, 
				  G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR);
	
	if (file_exist && !force) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_SAVE_FILE_EXIST,
			     _("%s"), real_filename);
		
		g_free (real_filename);

		return FALSE;
	}

	parser.doc = xmlNewDoc ("1.0");

	if (!mpp_write_project (&parser)) {
		g_warning ("Failed to save tree.");
	}

	ret = xmlSaveFormatFile (real_filename, parser.doc, 1);

	g_free (real_filename);
	
	g_hash_table_destroy (parser.task_hash);
	g_hash_table_destroy (parser.resource_hash);

	xmlFreeDoc (parser.doc);

	if (ret == -1) {
		g_set_error (error,
			     MRP_ERROR,
			     MRP_ERROR_SAVE_WRITE_FAILED,
			     _("Could not write xml file"));

		return FALSE;
	}
	
	return TRUE;
}


/***************
 * XML helpers.
 */


/* Get a value for a node either carried as an attibute or as
 * the content of a child.
 *
 * Returns xmlChar, must be xmlFreed.
 */
static gchar *
mpp_xml_get_value (xmlNodePtr node, const char *name)
{
	gchar      *val;
	xmlNodePtr  child;

	val = (gchar *) xmlGetProp (node, name);

	if (val != NULL) {
		return val;
	}

	child = node->children;

	while (child != NULL) {
		if (!strcmp (child->name, name)) {
		        /*
			 * !!! Inefficient, but ...
			 */
			val = xmlNodeGetContent(child);
			if (val != NULL) {
				return val;
			}
		}
		child = child->next;
	}

	return NULL;
}

/*
 * Get a value for a node either carried as an attibute or as
 * the content of a child.
 *
 * Returns a g_malloc'ed string.  Caller must free.
 */
static gchar *
mpp_xml_get_string (xmlNodePtr node, const char *name)
{
	char *ret, *val;

	g_return_val_if_fail (node != NULL, NULL);
	g_return_val_if_fail (name != NULL, NULL);

	val = mpp_xml_get_value (node, name);
	ret = g_strdup (val);
	xmlFree (val);

	return ret;
}

static mrptime
mpp_xml_get_date (xmlNodePtr node, const char *name)
{
	gchar   *val;
	mrptime  t;

	g_return_val_if_fail (node != NULL, MRP_TIME_INVALID);
	g_return_val_if_fail (name != NULL, MRP_TIME_INVALID);

	val = mpp_xml_get_value (node, name);

	t = mrp_time_from_string (val, NULL);
	
	xmlFree (val);

	return t;
}

static gint
mpp_xml_get_int (xmlNodePtr node, const char *name)
{
	gchar *val;
	gint   ret;

	g_return_val_if_fail (node != NULL, 0);
	g_return_val_if_fail (name != NULL, 0);

	val = mpp_xml_get_value (node, name);
	if (val == NULL) {
		return 0;
	}

	ret = atoi (val);
	
	xmlFree (val);

	return ret;
}

static gint
mpp_xml_get_int_with_default (xmlNodePtr node, const char *name, gint def)
{
	gchar *val;
	gint   ret;

	g_return_val_if_fail (node != NULL, def);
	g_return_val_if_fail (name != NULL, def);

	val = mpp_xml_get_value (node, name);
	if (val == NULL) {
		return def;
	}

	ret = atoi (val);
	
	xmlFree (val);

	return ret;
}

static gfloat
mpp_xml_get_float (xmlNodePtr node, const char *name)
{
	gchar *val;
	gfloat ret;

	g_return_val_if_fail (node != NULL, 0);
	g_return_val_if_fail (name != NULL, 0);

	val = mpp_xml_get_value (node, name);
	if (val == NULL) {
		return 0;
	}

	ret = g_ascii_strtod (val, NULL);
	
	xmlFree (val);

	return ret;
}

/* Search a child by name, if needed go down the tree to find it.
 */
static xmlNodePtr
mpp_xml_search_child (xmlNodePtr node, const gchar *name)
{
	xmlNodePtr ret;
	xmlNodePtr child;

	child = node->children;
	while (child != NULL) {
		if (!strcmp (child->name, name))
			return child;
		child = child->next;
	}
	child = node->children;
	while (child != NULL) {
		ret = mpp_xml_search_child (child, name);
		if (ret != NULL)
			return ret;
		child = child->next;
	}
	return NULL;
}

static void
mpp_xml_set_date (xmlNodePtr node, const gchar *prop, mrptime time)
{
	gchar *str;

	str = mrp_time_to_string (time);
	xmlSetProp (node, prop, str);
	g_free (str);
}

static void
mpp_xml_set_int (xmlNodePtr node, const gchar *prop, gint value)
{
	gchar *str;

	str = g_strdup_printf ("%d", value);
	xmlSetProp (node, prop, str);
	g_free (str);
}

static void
mpp_xml_set_float (xmlNodePtr node, const gchar *prop, gfloat value)
{
	gchar  buf[128];
	gchar *str;

	str = g_ascii_dtostr (buf, sizeof(buf) - 1, value);
	xmlSetProp (node, prop, str);
}
